"""Slices BI Images API."""

import itertools
from collections.abc import Iterator
from datetime import datetime
from typing import TYPE_CHECKING, Annotated

from httpx import Client, HTTPError
from msgspec import Meta, Struct
from msgspec.json import Decoder as JsonDecoder

from .exceptions import from_httpx_exception
from .pagination import PagedCollection
from .utils import json_encoder

if TYPE_CHECKING:
    from collections.abc import Collection


class VMFlavor(Struct, tag_field="flavor_type", tag="vm"):
    """VM Flavor."""

    id: str
    "TypeID uniquely identifying a resource"
    created_at: Annotated[datetime, Meta(tz=True)]
    "Timestamp when this flavor was created"

    friendly_name: str
    "User readable ID"
    description: str
    "Description of this flavor"

    disk_mb: int
    """The size of the root disk that will be created in MB.
    If 0 the root disk will be set to exactly the size of the image used to deploy the instance."""
    ram_mib: int
    "The amount of RAM in MiB"
    vcpus: int
    "The number of virtual CPUs that will be allocated to the server"

    owner_project: str | None = None
    "TypeID of the project owning the flavor, or None if this is a global project"


class BareMetalFlavor(Struct, tag_field="flavor_type", tag="baremetal"):
    """Bare metal flavor."""

    created_at: Annotated[datetime, Meta(tz=True)]
    "Timestamp when this flavor was created"
    id: str
    "TypeID uniquely identifying a resource"

    friendly_name: str
    "User readable ID"
    description: str
    "Description of this flavor"

    disk_mb: int
    """The size of the root disk that will be created in MB.
    If 0 the root disk will be set to exactly the size of the image used to deploy the instance."""
    ram_mib: int
    "The amount of RAM in MiB"
    cpu_cores: int
    "CPU cores (per CPU socket)"
    cpu_model: str
    "CPU model."
    cpu_sockets: int
    "CPU count (= CPU socket count)"
    cpu_threads: int
    "CPU Threads (per CPU core)"
    gpu_model: str | None
    network_interface_count: int
    "Number of test network interfaces present. (There is always at least the 'control' interface)"

    owner_project: str | None = None
    "TypeID of the project owning the flavor, or None if this is a global project"


AnyFlavor = VMFlavor | BareMetalFlavor
_flavor_decoder = JsonDecoder(AnyFlavor)


class VMFlavorCreate(Struct, tag_field="flavor_type", tag="vm"):
    """VM Flavor Create."""

    friendly_name: str
    "User readable ID"
    description: str
    "Description of this flavor"

    disk_mb: int
    """The size of the root disk that will be created in MB.
    If 0 the root disk will be set to exactly the size of the image used to deploy the instance."""
    ram_mib: int
    "The amount of RAM in MiB"
    vcpus: int
    "The number of virtual CPUs that will be allocated to the server"


class BareMetalFlavorCreate(Struct, tag_field="flavor_type", tag="baremetal"):
    """Bare metal flavor Create."""

    friendly_name: str
    "User readable ID"
    description: str
    "Description of this flavor"

    disk_mb: int
    """The size of the root disk that will be created in MB.
    If 0 the root disk will be set to exactly the size of the image used to deploy the instance."""
    ram_mib: int
    "The amount of RAM in MiB"
    cpu_cores: int
    "CPU cores (per CPU socket)"
    cpu_model: str
    "CPU model."
    cpu_sockets: int
    "CPU count (= CPU socket count)"
    cpu_threads: int
    "CPU Threads (per CPU core)"
    gpu_model: str | None
    network_interface_count: int
    "Number of test network interfaces present. (There is always at least the 'control' interface)"


AnyFlavorCreate = VMFlavorCreate | BareMetalFlavorCreate


class BiFlavorClient:
    """Basic Infrastructure Flavor API Client."""

    API_BASE_PATH = "/apis/bi.slices.eu/v1alpha2/"

    def __init__(self, client: Client, project_id: str) -> None:
        """Create Basic Infrastructur Flavor Client."""
        self._client = client

        self.project_id = project_id

        assert client.base_url is not None, "Expected base_url to be set"

    def get(self, id: str) -> AnyFlavor:
        """Get flavor.

        :param id: Flavor ID
        """

        # Doesn't matter if it is a project Flavor, the global endpoint will just redirect to it.
        response = self._client.get(f"{self.API_BASE_PATH}flavors/{id}")

        try:
            response.raise_for_status()
            return _flavor_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def create_global(self, flavor: AnyFlavorCreate) -> AnyFlavor:
        """Create a new global flavor. Admin only.

        :param flavor: Flavor creation object
        :return: Created flavor
        """
        response = self._client.post(f"{self.API_BASE_PATH}flavors", json=json_encoder.encode(flavor))

        try:
            response.raise_for_status()
            return _flavor_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def create_in_project(self, flavor: AnyFlavorCreate) -> AnyFlavor:
        """Create a new flavor in the current project.

        :param flavor: Flavor creation object
        :return: Created flavor
        """
        response = self._client.post(
            f"{self.API_BASE_PATH}projects/{self.project_id}/flavors", json=json_encoder.encode(flavor)
        )

        try:
            response.raise_for_status()
            return _flavor_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update(self, flavor: AnyFlavor) -> AnyFlavor:
        """Update flavor.

        :param flavor: Flavor
        :return: Updated flavor
        """
        response = self._client.put(f"{self.API_BASE_PATH}flavors/{id}", json=json_encoder.encode(flavor))

        try:
            response.raise_for_status()
            return _flavor_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def delete(self, flavor: AnyFlavor) -> AnyFlavor:
        """Delete flavor.

        :param flavor: Flavor
        :return: Deleted flavor
        """
        return self.delete_by_id(flavor.id)

    def delete_by_id(self, id: str) -> AnyFlavor:
        """Delete flavor.

        :param id: Flavor ID
        :return: Deleted flavor
        """
        response = self._client.delete(f"{self.API_BASE_PATH}flavors/{id}")

        try:
            response.raise_for_status()
            return _flavor_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list(self, include_global_flavors: bool = True, include_project_flavors: bool = True) -> Iterator[AnyFlavor]:
        """List all flavors."""

        res: list[Collection[AnyFlavor]] = []

        if include_global_flavors:
            res.append(PagedCollection("flavor", AnyFlavor, f"{self.API_BASE_PATH}flavors", self._client))  # pyright: ignore [reportArgumentType]
        if include_project_flavors:
            res.append(
                PagedCollection(
                    "flavor",
                    AnyFlavor,  # pyright: ignore [reportArgumentType]
                    f"{self.API_BASE_PATH}projects/{self.project_id}/flavors",
                    self._client,
                )
            )

        return itertools.chain.from_iterable(res)
