"""Slices BI Images API."""

from collections.abc import Collection
from datetime import datetime
from enum import Enum
from typing import Annotated

import msgspec.json
from httpx import Client, HTTPError
from msgspec import Meta, Struct
from msgspec.json import Decoder as JsonDecoder

from .exceptions import from_httpx_exception
from .pagination import PagedCollection


class DiskImageCapability(str, Enum):
    """Capabilities of a disk image."""

    PING = "ping"
    SSH = "ssh"
    CLOUD_MINUS_INIT = "cloud-init"
    CLOUD_MINUS_INIT_MINUS_PHONEHOME = "cloud-init-phonehome"


class DiskImageBase(Struct):
    """Disk Image Base."""

    friendly_name: str
    "User readable name"

    default_username: str | None = None

    location: str | None = None
    min_disk_mb: int | None = None
    min_ram_mib: int | None = None
    tags: list[str] | None = None
    cluster_id: str | None = None
    capabilities: list[DiskImageCapability] | None = None

    short_description: str | None = None
    long_description: str | None = None


class DiskImage(Struct):
    """Disk Image."""

    id: str
    "TypeID uniquely identifying a resource"

    created_at: Annotated[datetime, Meta(tz=True)]
    "Timestamp when this disk image was created"

    friendly_name: str
    "User readable name"

    default_username: str | None = None

    location: str | None = None
    min_disk_mb: int | None = None
    min_ram_mib: int | None = None
    tags: list[str] | None = None
    cluster_id: str | None = None
    capabilities: list[DiskImageCapability] | None = None

    short_description: str | None = None
    long_description: str | None = None


class DiskImageCreate(DiskImageBase):
    """Disk Image Create."""


class DiskImageUpdate(DiskImageBase):
    """Disk Image Update."""


_disk_image_decoder = JsonDecoder(DiskImage)


class BiImageClient:
    """Basic Infrastructure Images API Client."""

    API_BASE_PATH = "/apis/bi.slices.eu/v1alpha2/images/"

    def __init__(self, client: Client) -> None:
        """Create Image Client."""
        self._client = client

        assert client.base_url is not None, "Expected base_url to be set"

    def create(
        self,
        friendly_name: str,
        default_username: str | None = None,
        location: str | None = None,
        min_disk_mb: int | None = None,
        min_ram_mib: int | None = None,
        tags: list[str] | None = None,
        cluster_id: str | None = None,
        capabilities: list[DiskImageCapability] | None = None,
        short_description: str | None = None,
        long_description: str | None = None,
    ) -> DiskImage:
        """Create disk image."""
        response = self._client.post(
            self.API_BASE_PATH,
            headers={"Content-Type": "application/json"},
            content=msgspec.json.encode(
                DiskImageCreate(
                    friendly_name=friendly_name,
                    default_username=default_username,
                    location=location,
                    min_disk_mb=min_disk_mb,
                    min_ram_mib=min_ram_mib,
                    tags=tags,
                    cluster_id=cluster_id,
                    capabilities=capabilities,
                    short_description=short_description,
                    long_description=long_description,
                )
            ),
        )

        try:
            response.raise_for_status()
            return _disk_image_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def get(self, id: str) -> DiskImage:
        """Get disk image."""
        response = self._client.get(f"{self.API_BASE_PATH}{id}")

        try:
            response.raise_for_status()
            return _disk_image_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def update(
        self,
        id: str,
        friendly_name: str,
        default_username: str | None = None,
        location: str | None = None,
        min_disk_mb: int | None = None,
        min_ram_mib: int | None = None,
        tags: list[str] | None = None,
        cluster_id: str | None = None,
    ) -> DiskImage:
        """Update disk image."""
        response = self._client.put(
            f"{self.API_BASE_PATH}{id}",
            headers={"Content-Type": "application/json"},
            content=msgspec.json.encode(
                DiskImageUpdate(
                    friendly_name=friendly_name,
                    default_username=default_username,
                    location=location,
                    min_disk_mb=min_disk_mb,
                    min_ram_mib=min_ram_mib,
                    tags=tags,
                    cluster_id=cluster_id,
                )
            ),
        )

        try:
            response.raise_for_status()
            return _disk_image_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def delete(self, disk_image: DiskImage) -> DiskImage:
        """Delete disk image."""
        return self.delete_by_id(disk_image.id)

    def delete_by_id(self, id: str) -> DiskImage:
        """Delete disk image."""
        response = self._client.delete(f"{self.API_BASE_PATH}{id}")

        try:
            response.raise_for_status()
            return _disk_image_decoder.decode(response.content)
        except HTTPError as err:
            raise from_httpx_exception(err) from err

    def list(self) -> Collection[DiskImage]:
        """List experiments."""

        return PagedCollection("disk_image", DiskImage, self.API_BASE_PATH, self._client)
