"""Slices Client."""

import importlib.metadata
from datetime import datetime, timedelta
from functools import cached_property
from uuid import uuid4

import httpx
import jwt
from pytz import UTC

from .bi import BiClient, BiExperimentClient
from .config import (
    MINIMUM_LIFETIME_FOR_FRESH_TOKEN,
    ApiClientsConfig,
    ApiEntry,
    ensure_auth_token_is_fresh,
    load_api_clients_config,
    load_auth_token,
    parse_auth_token,
)
from .exceptions import ApiException
from .experiment import ExperimentClient
from .oauth2.token import AuthJWT, OAuthToken
from .oauth2.token_exchange import ExchangedToken, get_access_token_for_audience
from .project import ProjectClient
from .pubkey import PublicKeyClient
from .user import UserClient


class NotLoggedInException(Exception):
    """Not logged in."""


def add_request_id_header(request):
    """Add Request ID header."""
    request.headers["X-Request-ID"] = str(uuid4())
    return request


def _get_http_user_agent():
    """Get HTTP User Agent."""
    return f"slices-bi-client/{importlib.metadata.version('slices_bi_client')}"


class OAuthHttpxAuth(httpx.Auth):
    """OAuth HTTPX Auth."""

    _token_for_audience: ExchangedToken | None = None
    _token_expires_at: datetime | None = None

    def __init__(self, slices_client: "SlicesClient", target_audience: str, project_id: str | None = None) -> None:
        """Create HTTPX Auth provider that sets the appropriate Authorization-header."""
        self.slices_client = slices_client
        self.target_audience = target_audience
        self.project_id = project_id

    def auth_flow(self, request):
        """Auth flow."""

        if not self._token_expires_at or (datetime.now() > self._token_expires_at + MINIMUM_LIFETIME_FOR_FRESH_TOKEN):
            # This token is scoped to the CLI
            auth_token = self.slices_client._get_fresh_auth_token()

            # So we need to exchange it for a token that is scoped to the target audience
            self._token_for_audience = get_access_token_for_audience(auth_token, self.target_audience, self.project_id)
            self._token_expires_at = datetime.now() + timedelta(seconds=self._token_for_audience.expires_in)

        assert self._token_for_audience

        request.headers["Authorization"] = f"Bearer {self._token_for_audience.access_token}"

        # Add a request ID header
        # While not directly related to authentication, we do it here because
        # it's an easy location to mutate the request.
        add_request_id_header(request)
        yield request


class SlicesClient:
    """Slices Client."""

    _auth_token: OAuthToken | None
    _api_clients_config: ApiClientsConfig

    def __init__(self, config: ApiClientsConfig | None = None) -> None:
        """Create Slices Client."""
        self._auth_token = load_auth_token()
        self._api_clients_config = config or load_api_clients_config()

    def _create_httpx_client(
        self, *, base_url: str, audience: str, project_id: str | None = None, experiment_id: str | None = None
    ) -> httpx.Client:
        auth = OAuthHttpxAuth(self, audience, project_id)
        if experiment_id:
            if not project_id:
                raise ValueError("project_id must be set when experiment_id is set")
            auth = ExperimentJwtHttpxAuth(self, experiment_id, project_id)

        return httpx.Client(
            base_url=base_url,
            auth=auth,
            headers={"User-Agent": _get_http_user_agent(), "Accept": "application/json"},
            follow_redirects=True,
            max_redirects=5,
            timeout=httpx.Timeout(timeout=5.0),
        )

    def _create_httpx_client_for_api_entry(
        self,
        api_entry: ApiEntry,
        project_id: str | None = None,
        experiment_id: str | None = None,
    ) -> httpx.Client:
        return self._create_httpx_client(
            base_url=api_entry.base_url, audience=api_entry.audience, project_id=project_id, experiment_id=experiment_id
        )

    def _get_fresh_auth_token(self):
        if not self._auth_token:
            raise NotLoggedInException("Not logged in")

        self._auth_token = ensure_auth_token_is_fresh(self._auth_token)
        return self._auth_token

    @property
    def user_info(self) -> AuthJWT | None:
        """Auth Token."""
        return self._auth_token and parse_auth_token(self._auth_token)

    @cached_property
    def project(self):
        """Project API Client."""
        return ProjectClient(self._create_httpx_client_for_api_entry(self._api_clients_config.project_authority))

    @cached_property
    def user(self):
        """User API Client."""
        return UserClient(self._create_httpx_client_for_api_entry(self._api_clients_config.user_authority))

    @cached_property
    def experiment(self):
        """Experiment API Client."""
        return ExperimentClient(
            self._create_httpx_client_for_api_entry(self._api_clients_config.experiment_authority),
            project_client=self.project,
        )

    @cached_property
    def public_key(self):
        """Public Key API Client."""
        return PublicKeyClient(self._create_httpx_client_for_api_entry(self._api_clients_config.pubkey))

    def for_project(self, project_id: str) -> "SlicesProjectClient":
        """Get Slices Project Client."""
        return SlicesProjectClient(project_id, self)

    def for_experiment(self, experiment_id: str, project_id: str) -> "SlicesExperimentClient":
        """Get Slices Experiment Client."""
        return SlicesExperimentClient(experiment_id, project_id, self)


class SlicesProjectClient:
    """Slices client for a specific project."""

    def __init__(self, project_id: str, generic_client: SlicesClient) -> None:
        """Create Slices Project Client."""
        super().__init__()
        self.project_id = project_id
        self._generic_client = generic_client

    def bi(self, site_id: str) -> BiClient:
        """Get Basic Infrastructure API Client for a specific site."""

        if site_id == "bi-orchestrator":
            return BiClient(
                self._generic_client._create_httpx_client_for_api_entry(
                    self._generic_client._api_clients_config.orchestrator, self.project_id
                ),
                self.project_id,
            )

        bi_api_entry = self._generic_client._api_clients_config.bi.get(site_id)

        if not bi_api_entry:
            raise ValueError(f"Unknown Site ID: {site_id}")

        return BiClient(
            self._generic_client._create_httpx_client_for_api_entry(bi_api_entry, self.project_id), self.project_id
        )

    def for_experiment(self, experiment_id: str) -> "SlicesExperimentClient":
        """Get Slices Experiment Client."""
        return SlicesExperimentClient(experiment_id, self.project_id, self._generic_client)


class ExperimentJwtHttpxAuth(httpx.Auth):
    """HTTPX Auth implementation which fetches a JWT for a specific experiment."""

    _token: str | None = None
    _token_expires_at: datetime | None = None

    def __init__(self, slices_client: SlicesClient, experiment_id: str, project_id: str) -> None:
        """Experiment JWT Authentication for HTTPX."""
        super().__init__()
        self.slices_client = slices_client
        self.experiment_id = experiment_id
        self.project_id = project_id

    def get_experiment_jwt(self, experiment_id: str) -> str:
        """Get Experiment JWT."""
        return self.slices_client.experiment.get_jwt(experiment_id)

    def auth_flow(self, request):
        """Auth flow."""

        if not self._token_expires_at or (datetime.now() > self._token_expires_at + MINIMUM_LIFETIME_FOR_FRESH_TOKEN):
            # So we need to exchange it for a token that is scoped to the target audience
            self._token = self.slices_client.experiment.get_jwt(self.experiment_id)

            decoded_token = jwt.decode(self._token, options={"verify_signature": False})

            if (exp_proj := decoded_token.get("proj")) != self.project_id:
                raise ApiException(
                    f"'{self.experiment_id}' is an experiment in project '{exp_proj}', "
                    f"but was expecting it to be from project '{self.project_id}'"
                )

            self._token_expires_at = datetime.fromtimestamp(decoded_token["exp"], UTC)

        assert self._token

        request.headers["Authorization"] = f"Bearer {self._token}"

        yield request


class SlicesExperimentClient:
    """Slices client for a specific experiment."""

    def __init__(self, experiment_id: str, project_id: str, generic_client: SlicesClient) -> None:
        """Create Slices Project Client."""
        super().__init__()
        self.experiment_id = experiment_id
        self.project_id = project_id
        self._generic_client = generic_client

    def bi(self, site_id: str) -> BiExperimentClient:
        """Get Basic Infrastructure API Client for a specific site."""

        bi_api_entry = (
            self._generic_client._api_clients_config.bi.get(site_id)
            if site_id != "bi-orchestrator"
            else self._generic_client._api_clients_config.orchestrator
        )

        if not bi_api_entry:
            raise ValueError(f"Unknown Site ID: {site_id}")

        return BiExperimentClient(
            self._generic_client._create_httpx_client_for_api_entry(bi_api_entry, self.project_id, self.experiment_id),
            experiment_id=self.experiment_id,
            project_id=self.project_id,
            project_client=self._generic_client.project,
        )

    @cached_property
    def bi_orchestrator(self) -> BiExperimentClient:
        """Get Basic Infrastructure API Client for the orchestrator site."""
        bi_api_entry = self._generic_client._api_clients_config.orchestrator

        return BiExperimentClient(
            self._generic_client._create_httpx_client_for_api_entry(bi_api_entry, self.project_id, self.experiment_id),
            experiment_id=self.experiment_id,
            project_id=self.project_id,
            project_client=self._generic_client.project,
        )
