"""CLI configuration management."""

import logging
import os
import subprocess
from datetime import datetime, timedelta
from functools import cache
from pathlib import Path
from typing import Literal

import jwt
import msgspec.json
from msgspec import Struct
from pytz import UTC
from tomlkit import TOMLDocument, table
from tomlkit.exceptions import TOMLKitError
from tomlkit.items import Table
from tomlkit.toml_file import TOMLFile

from .const import AUTH_FILENAME, CONFIG_DIR, JWT_SUPPORTED_ALGORITHMS, SETTINGS_FILENAME
from .exceptions import AuthenticationError, ConfigurationError
from .oauth2.config import settings
from .oauth2.flow import RefreshFlow
from .oauth2.token import AuthJWT, OAuthToken

logger = logging.getLogger(__name__)

config_dir_location = Path.home() / CONFIG_DIR
auth_file_location = config_dir_location / AUTH_FILENAME
settings_file_location = config_dir_location / SETTINGS_FILENAME

jwks_client = jwt.PyJWKClient(settings.jwks_url)


class AuthConfig(Struct):
    """Auth config file contents."""

    version: Literal["0.1"]
    oauth_token: OAuthToken


def save_auth_token(auth_token: OAuthToken):
    """Save authentication token."""

    auth_config: AuthConfig = AuthConfig(
        version="0.1",
        oauth_token=auth_token,
    )

    config_dir_location.mkdir(exist_ok=True)
    with auth_file_location.open("wb") as f:
        f.write(msgspec.json.encode(auth_config))

    auth_file_location.chmod(0o600)  # set secure file permissions


def parse_auth_token(auth_token: OAuthToken) -> AuthJWT:
    """Parse auth token into an AuthJWT object."""
    jwt_token = auth_token.access_token
    signing_key = jwks_client.get_signing_key_from_jwt(jwt_token)

    return jwt.decode(
        jwt_token,
        signing_key.key,
        algorithms=JWT_SUPPORTED_ALGORITHMS,
        audience=settings.audience,
        leeway=15,
    )


def delete_auth_token():
    """Delete authentication token."""
    auth_file_location.unlink(missing_ok=True)


def refresh_auth_token(refresh_token: str) -> OAuthToken:
    """Refresh the authentication token."""

    new_auth_token = RefreshFlow().perform_refresh(refresh_token)

    save_auth_token(new_auth_token)

    return new_auth_token


MINIMUM_LIFETIME_FOR_FRESH_TOKEN = timedelta(minutes=1)


def ensure_auth_token_is_fresh(auth_token: OAuthToken) -> OAuthToken:
    """Ensure the authentication token is fresh."""

    if datetime.fromtimestamp(auth_token.expires_at, UTC) < (datetime.now(UTC) - MINIMUM_LIFETIME_FOR_FRESH_TOKEN):
        if auth_token.refresh_token:
            return refresh_auth_token(auth_token.refresh_token)

        raise AuthenticationError("Token expired and no refresh token available")

    return auth_token


def load_auth_token() -> OAuthToken | None:
    """Load authentication token from file."""

    # Check if we have an external token provider
    if auth_token_provider := os.environ.get("SLICES_AUTH_TOKEN_PROVIDER"):
        try:
            return msgspec.json.decode(subprocess.check_output(auth_token_provider), type=OAuthToken)  # noqa: S603
        except subprocess.CalledProcessError as e:
            raise ConfigurationError("Could not retrieve authentication token from external provider") from e
    if not auth_file_location.exists():
        return None

    with Path(auth_file_location).open() as f:
        auth_file_contents = msgspec.json.decode(f.read(), type=AuthConfig)

    if not auth_file_contents:
        logger.error("Could not load authentication token")
        return None

    if auth_file_contents.version != "0.1":
        logger.warning("Unexpected auth file version: %s", auth_file_contents.version)

    auth_token: OAuthToken = auth_file_contents.oauth_token

    if auth_token.refresh_token and datetime.fromtimestamp(auth_token.expires_at) < datetime.now():
        auth_token = refresh_auth_token(auth_token.refresh_token)

    return auth_token


@cache
def _load_settings() -> TOMLDocument:
    config_dir_location.mkdir(exist_ok=True)

    try:
        if settings_file_location.exists():
            settings_file = TOMLFile(settings_file_location)
            return settings_file.read()
    except TOMLKitError as e:
        raise ConfigurationError(f"Could not read settings file {settings_file_location}") from e

    return TOMLDocument()


def _save_settings(config: TOMLDocument):
    config_dir_location.mkdir(exist_ok=True)
    settings_file = TOMLFile(settings_file_location)

    settings_file.write(config)


def set_config_string(name: str, value: str | None, section: str | None = None):
    """Set config value."""

    settings = _load_settings()

    if section:
        if section not in settings:
            section_table = table()
            settings.add(section, section_table)
        else:
            section_table = settings[section]
            if not isinstance(section_table, Table):
                raise ValueError(f"Expected '{section}' to be a table.")

        if value:
            section_table[name] = value
        else:
            del section_table[name]
    else:
        settings[name] = value

    _save_settings(settings)


def get_config_string(name: str, default=None, section: str | None = None) -> str | None:
    """Get config value."""

    settings = _load_settings()

    if section:
        if section in settings:
            section_item = settings[section]
            if not isinstance(section_item, Table):
                raise ValueError(f"Expected {section} to be a table.")
            return section_item.get(name, default)
        return default
    return settings.get(name, default)


class ApiEntry(Struct):
    """Information needed to connect to API."""

    base_url: str
    audience: str


class ApiClientsConfig(Struct):
    """Configuration model used to create api client objects."""

    project_authority: ApiEntry
    user_authority: ApiEntry
    experiment_authority: ApiEntry
    pubkey: ApiEntry
    bi: dict[str, ApiEntry]
    orchestrator: ApiEntry


@cache
def load_api_clients_config(config_location: Path | None = None) -> ApiClientsConfig:
    """Load API clients configuration."""

    if not config_location:
        config_location = Path(__file__).with_name("config.json")

    with config_location.open() as f:
        return msgspec.json.decode(f.read(), type=ApiClientsConfig)
