"""Token Exchange."""

import httpx
from msgspec import Struct
from msgspec.json import Decoder as JsonDecoder

from .config import settings
from .token import OAuthToken


class ExchangedToken(Struct):
    """Response from token exchange."""

    access_token: str
    expires_in: int
    issued_token_type: str
    scope: str
    token_type: str


token_decoder = JsonDecoder(ExchangedToken)


def get_access_token_for_audience(
    auth_token: OAuthToken, audience: str, project_id: str | None = None
) -> ExchangedToken:
    """Return JWT access token of logged in user for target resource."""
    scope = auth_token.scope
    if project_id:
        scope = f"{scope} project:{project_id}"

    data = {
        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
        "subject_token": auth_token.access_token,
        "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
        "client_id": settings.client_id,
        "client_secret": settings.client_secret,
        "requested_token_type": "urn:ietf:params:oauth:token-type:access_token",
        "audience": audience,
        "scope": scope,
    }

    response = httpx.post(
        url=settings.token_endpoint_url,
        data=data,
        timeout=10,
    )
    response.raise_for_status()
    return token_decoder.decode(response.content)
